package net.wanji.web.service.scheme.impl;

import net.wanji.web.common.enums.LightsAddressEnum;
import net.wanji.web.dto.CrossIdDTO;
import net.wanji.web.dto.DirListElement;
import net.wanji.web.dto.LaneListElement;
import net.wanji.web.dto.LedConfigListElement;
import net.wanji.web.dto.SaveLaneInfoDTO;
import net.wanji.web.mapper.scheme.CrossDirInfoMapper;
import net.wanji.web.mapper.scheme.CrossLaneLightsMapper;
import net.wanji.web.mapper.scheme.CrossLightsMapper;
import net.wanji.web.mapper.scheme.CrossPhaseLightsMapper;
import net.wanji.web.mapper.scheme.LaneInfoMapper;
import net.wanji.web.mapper.scheme.LaneSegmentMapper;
import net.wanji.web.po.scheme.CrossDirInfoPO;
import net.wanji.web.po.scheme.CrossLaneLightsPO;
import net.wanji.web.po.scheme.CrossLightsPO;
import net.wanji.web.po.scheme.LaneInfoPO;
import net.wanji.web.service.scheme.CrossConfigService;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @author Kent HAN
 * @date 2022/12/20 10:27
 */
@Service
public class CrossConfigServiceImpl implements CrossConfigService {
    private final CrossDirInfoMapper crossDirInfoMapper;
    private final LaneInfoMapper laneInfoMapper;
    private final CrossLaneLightsMapper crossLaneLightsMapper;
    private final LaneSegmentMapper laneSegmentMapper;
    private final CrossLightsMapper crossLightsMapper;
    // 灯组位置与放行方向对应关系
    private static final Map<Integer, Integer> ADDRESS_DIR_MAP = new HashMap<>(8);
    static {
        ADDRESS_DIR_MAP.put(1, 5);
        ADDRESS_DIR_MAP.put(2, 6);
        ADDRESS_DIR_MAP.put(3, 7);
        ADDRESS_DIR_MAP.put(4, 8);
        ADDRESS_DIR_MAP.put(5, 1);
        ADDRESS_DIR_MAP.put(6, 2);
        ADDRESS_DIR_MAP.put(7, 3);
        ADDRESS_DIR_MAP.put(8, 4);
    }

    // 放行方向与灯组位置对应关系
    private static final Map<Integer, Integer> DIR_ADDRESS_MAP = new HashMap<>(8);
    static {
        DIR_ADDRESS_MAP.put(5, 1);
        DIR_ADDRESS_MAP.put(6, 2);
        DIR_ADDRESS_MAP.put(7, 3);
        DIR_ADDRESS_MAP.put(8, 4);
        DIR_ADDRESS_MAP.put(1, 5);
        DIR_ADDRESS_MAP.put(2, 6);
        DIR_ADDRESS_MAP.put(3, 7);
        DIR_ADDRESS_MAP.put(4, 8);
    }

    // 机动车灯数据库类型与前端类型对应关系
    private static final Map<Integer, Integer> MOTOR_VEH_MAP = new HashMap<>(5);
    static {
        MOTOR_VEH_MAP.put(1, 1);
        MOTOR_VEH_MAP.put(2, 3);
        MOTOR_VEH_MAP.put(3, 4);
        MOTOR_VEH_MAP.put(4, 5);
        MOTOR_VEH_MAP.put(5, 6);
    }
    // 非机动车灯数据库类型与前端类型对应关系
    private static final Map<Integer, Integer> NON_MOTOR_VEH_MAP = new HashMap<>(5);
    static {
        MOTOR_VEH_MAP.put(10, 2);
        MOTOR_VEH_MAP.put(11, 3);
        MOTOR_VEH_MAP.put(12, 4);
        MOTOR_VEH_MAP.put(13, 5);
        MOTOR_VEH_MAP.put(14, 6);
    }

    private final CrossPhaseLightsMapper crossPhaseLightsMapper;

    public CrossConfigServiceImpl(CrossDirInfoMapper crossDirInfoMapper, LaneInfoMapper laneInfoMapper, CrossLaneLightsMapper crossLaneLightsMapper, LaneSegmentMapper laneSegmentMapper, CrossLightsMapper crossLightsMapper, CrossPhaseLightsMapper crossPhaseLightsMapper) {
        this.crossDirInfoMapper = crossDirInfoMapper;
        this.laneInfoMapper = laneInfoMapper;
        this.crossLaneLightsMapper = crossLaneLightsMapper;
        this.laneSegmentMapper = laneSegmentMapper;
        this.crossLightsMapper = crossLightsMapper;
        this.crossPhaseLightsMapper = crossPhaseLightsMapper;
    }

    @Override
    @Transactional
    public void saveLaneInfo(SaveLaneInfoDTO saveLaneInfoDTO) {
        List<LedConfigListElement> ledConfigList = saveLaneInfoDTO.getLedConfigList();
        if (ledConfigList == null || ledConfigList.size() == 0) {
            // 保存渠化数据
            handleLaneData(saveLaneInfoDTO);
        } else {
            // 前端传过来的所有灯组代码列表
            List<String> lightCodeList = new ArrayList<>();
            // 保存灯组设置
            handleLightsData(saveLaneInfoDTO, ledConfigList, lightCodeList);
            // 保存车道配置（灯组-车道关系表）
            handleLaneConfig(saveLaneInfoDTO, lightCodeList);
            if (lightCodeList.size() != 0) {
                throw new RuntimeException("存在未配置的灯组号");
            }
        }
    }

    private void handleLaneConfig(SaveLaneInfoDTO saveLaneInfoDTO, List<String> lightCodeList) {
        String crossId = saveLaneInfoDTO.getCrossId();
        List<DirListElement> dirList = saveLaneInfoDTO.getDirList();
        List<CrossLaneLightsPO> crossLaneLightsPOList = new ArrayList<>();
        for (DirListElement dirListElement : dirList) {
            List<LaneListElement> laneList = dirListElement.getLaneList();
            Integer dir = dirListElement.getDir();
            for (LaneListElement laneListElement : laneList) {
                // 根据路口ID、方向、车道代码获取车道ID
                String code = laneListElement.getName(); // 车道代码
                LaneInfoPO laneInfoPO = laneInfoMapper.selectByCrossIdDirCode(crossId, dir, code);
                String laneId = laneInfoPO.getId();
                CrossLaneLightsPO crossLaneLightsPO = new CrossLaneLightsPO();
                crossLaneLightsPO.setLaneId(laneId);
                crossLaneLightsPO.setCrossId(crossId);
                // 根据路口ID和灯组代码获取灯组ID
                String ledNum = laneListElement.getLedNum();
                if (ledNum != null) {
                    CrossLightsPO crossLightsPO = crossLightsMapper.selectByCrossIdAndLedNum(crossId, ledNum);
                    Integer lightsId = crossLightsPO.getId();
                    crossLaneLightsPO.setLightsId(lightsId);
                    // 用于判断灯组号是否都配置完成
                    lightCodeList.remove(ledNum);
                }
                crossLaneLightsPOList.add(crossLaneLightsPO);
            }
        }
        crossLaneLightsMapper.deleteByCrossId(crossId);
        crossLaneLightsMapper.insertBatch(crossLaneLightsPOList);
    }

    private void handleLightsData(SaveLaneInfoDTO saveLaneInfoDTO, List<LedConfigListElement> ledConfigList,
                                  List<String> lightCodeList) {
        String crossId = saveLaneInfoDTO.getCrossId();
        List<CrossLightsPO> crossLightsPOList = new ArrayList<>();
        for (LedConfigListElement ledConfigListElement : ledConfigList) {
            String lightCode = ledConfigListElement.getCode();
            lightCodeList.add(lightCode);
            CrossLightsPO crossLightsPO = new CrossLightsPO();
            crossLightsPO.setLightsNo(lightCode);
            // 灯组类型
            setType(ledConfigListElement, crossLightsPO);
            // 灯组放行方向
            String address = ledConfigListElement.getAddress();
            Integer addressCode = LightsAddressEnum.getCodeByMsg(address);
            if (addressCode < 10) { // 非行人灯
                Integer dir = ADDRESS_DIR_MAP.get(addressCode);
                crossLightsPO.setDir(dir);
            } else { // 行人灯
                String firstDigit = address.toString().substring(0, 1);
                crossLightsPO.setDir(Integer.parseInt(firstDigit));
            }
            crossLightsPO.setSort(ledConfigListElement.getOrder());
            crossLightsPO.setCrossId(crossId);
            crossLightsPOList.add(crossLightsPO);
        }
        crossLightsMapper.deleteByCrossId(crossId);
        crossLightsMapper.insertBatch(crossLightsPOList);
    }

    private static void setType(LedConfigListElement ledConfigListElement, CrossLightsPO crossLightsPO) {
        Integer type = ledConfigListElement.getType();
        Integer signalType = ledConfigListElement.getSignalType();
        String address = ledConfigListElement.getAddress();
        if (type == 1) { // 机动车灯
            if (signalType == 1) {
                // 圆饼灯
                crossLightsPO.setType(1);
            } else if (signalType == 3) {
                // 左转灯
                crossLightsPO.setType(2);
            } else if (signalType == 4) {
                // 直行灯
                crossLightsPO.setType(3);
            } else if (signalType == 5) {
                // 右转灯
                crossLightsPO.setType(4);
            } else if (signalType == 6) {
                // 掉头灯
                crossLightsPO.setType(5);
            }
        } else if (type == 2) { // 非机动车灯
            if (signalType == 2) {
                // 无转向的非机动车灯
                crossLightsPO.setType(10);
            } else if (signalType == 3) {
                // 左转非机动车灯
                crossLightsPO.setType(11);
            } else if (signalType == 4) {
                // 直行非机动车灯
                crossLightsPO.setType(12);
            } else if (signalType == 5) {
                // 右转非机动车灯
                crossLightsPO.setType(13);
            } else if (signalType == 6) {
                // 掉头非机动车灯
                crossLightsPO.setType(14);
            }
        } else if (type == 3) { // 公交专用
            if (signalType == 7) {
                crossLightsPO.setType(30);
            }
        } else if (type == 4) { // 行人
            Integer addressCode = LightsAddressEnum.getCodeByMsg(address);
            if (addressCode >= 10 && addressCode <= 99) { // 两位数为一次过街
                crossLightsPO.setType(20);
            } else if (addressCode >= 100 && addressCode % 10 == 1) { // 三位数为二次过街，个位是1表示出口
                crossLightsPO.setType(22);
            } else if (addressCode >= 100 && addressCode % 10 == 2) { // 三位数为二次过街，个位是2表示进口
                crossLightsPO.setType(21);
            }
        }
    }

    private void handleLaneData(SaveLaneInfoDTO saveLaneInfoDTO) {
        String crossId = saveLaneInfoDTO.getCrossId();
        List<DirListElement> dirList = saveLaneInfoDTO.getDirList();
        for (DirListElement dirListElement : dirList) {
            // 获取数据库中已有数据
            Integer dir = dirListElement.getDir();
            List<LaneInfoPO> laneInfoPOList = laneInfoMapper.selectByCrossIdAndDir(crossId, dir);
            // 按车道代码排序
            List<LaneListElement> laneListFromClient = dirListElement.getLaneList();
            laneListFromClient.sort((x,y) -> String.CASE_INSENSITIVE_ORDER.compare(x.getName(), y.getName()));
            List<LaneListElement> laneListFromDb = buildLaneListFromDb(laneInfoPOList);
            laneListFromDb.sort((x,y) -> String.CASE_INSENSITIVE_ORDER.compare(x.getName(), y.getName()));
            // 比较入参数据与已有数据
            if (!laneListFromClient.equals(laneListFromDb)) {
                // 删除车道表数据
                laneInfoMapper.deleteByCrossIdAndDir(crossId, dir);
                // 删除路口方向表数据
                crossDirInfoMapper.deleteInByDirType(crossId, dir);
                List<String> laneIds = getLaneIds(laneInfoPOList);
                if (laneIds.size()!=0){
                    // 删除灯组-车道关系表数据
                    crossLaneLightsMapper.deleteByLaneIds(laneIds);
                    // 删除车道状态表数据
                    laneSegmentMapper.deleteByLaneIds(laneIds);
                }
                // 插入进口信息
                List<LaneInfoPO> laneInfoPOListForInsert = getLaneInfoPOListForInsert(
                        crossId, dirListElement, dir, laneListFromClient);
                laneInfoMapper.insertBatch(laneInfoPOListForInsert);
                // 插入路口方向
                CrossDirInfoPO crossDirInfoPO = getCrossDirInfoPO(crossId, dirListElement);
                crossDirInfoMapper.insertOne(crossDirInfoPO);
            }
            // 更新是否有行人道
            String id = getId(crossId, dirListElement);
            crossDirInfoMapper.updateIsPedestrian(dirListElement.getIsPersonCross(), id);
        }
    }

    @Override
    public SaveLaneInfoDTO listLaneInfo(CrossIdDTO crossIdDTO) {
        String crossId = crossIdDTO.getCrossId();
        SaveLaneInfoDTO saveLaneInfoDTO = new SaveLaneInfoDTO();
        saveLaneInfoDTO.setCrossId(crossId);
        // 构造dirList
        buildDirList(crossId, saveLaneInfoDTO);
        // 构造ledConfigList
        buildLedConfigList(crossId, saveLaneInfoDTO);

        return saveLaneInfoDTO;
    }

    private void buildLedConfigList(String crossId, SaveLaneInfoDTO saveLaneInfoDTO) {
        saveLaneInfoDTO.setLedConfigList(new ArrayList<>());
        List<LedConfigListElement> ledConfigList = saveLaneInfoDTO.getLedConfigList();
        // 获取灯组数据
        List<CrossLightsPO> crossLightsPOList = crossLightsMapper.selectByCrossId(crossId);
        // 赋值给ledConfigList
        for (CrossLightsPO crossLightsPO : crossLightsPOList) {
            LedConfigListElement ledConfigListElement = new LedConfigListElement();
            ledConfigListElement.setId(crossLightsPO.getId());
            ledConfigListElement.setOrder(crossLightsPO.getSort());
            ledConfigListElement.setCode(crossLightsPO.getLightsNo());
            // 处理type
            Integer type = crossLightsPO.getType();
            if (type > 0 && type < 10) { // 机动车灯
                ledConfigListElement.setType(1);
                Integer signalType = MOTOR_VEH_MAP.get(type);
                ledConfigListElement.setSignalType(signalType);
            } else if (type >= 10 && type < 20) { // 非机动车灯
                ledConfigListElement.setType(2);
                Integer signalType = NON_MOTOR_VEH_MAP.get(type);
                ledConfigListElement.setSignalType(signalType);
            } else if (type >= 20 && type < 30) { // 行人灯
                ledConfigListElement.setType(4);
                ledConfigListElement.setSignalType(8);
            } else if (type == 30) { // 公交专用灯
                ledConfigListElement.setType(3);
                ledConfigListElement.setSignalType(7);
            }
            // 灯组放行方向转换为灯组位置
            Integer dir = crossLightsPO.getDir();
            Integer addressCode = DIR_ADDRESS_MAP.get(dir);
            if (type < 20 || type >= 30) { // 非行人灯
                String address = LightsAddressEnum.getMsgByCode(addressCode);
                ledConfigListElement.setAddress(address);
            } else if (type == 20) { // 一次过街
                String s = "" + dir + 1;
                setLightAddress(ledConfigListElement, s);
            } else if (type == 22) { // 二次过街-出口
                String s = "" + dir + 21;
                setLightAddress(ledConfigListElement, s);
            } else if (type == 21) { // 二次过街-进口
                String s = "" + dir + 22;
                setLightAddress(ledConfigListElement, s);
            }
            ledConfigList.add(ledConfigListElement);
        }
        // 根据灯组序号排序
        ledConfigList.sort(Comparator.comparing(LedConfigListElement::getOrder));
    }

    private static void setLightAddress(LedConfigListElement ledConfigListElement, String s) {
        int i = Integer.parseInt(s);
        String address = LightsAddressEnum.getMsgByCode(i);
        ledConfigListElement.setAddress(address);
    }

    private void buildDirList(String crossId, SaveLaneInfoDTO saveLaneInfoDTO) {
        saveLaneInfoDTO.setDirList(new ArrayList<>());
        List<LaneInfoPO> laneInfoPOList = laneInfoMapper.selectBycrossId(crossId);
        Map<Integer, List<LaneInfoPO>> collect = laneInfoPOList.stream()
                .collect(Collectors.groupingBy(LaneInfoPO::getDir));
        for (Map.Entry<Integer, List<LaneInfoPO>> entry : collect.entrySet()) {
            Integer key = entry.getKey();
            DirListElement dirListElement = new DirListElement();
            dirListElement.setDir(key);
            // 获取是否有行人道
            CrossDirInfoPO crossDirInfoPO = crossDirInfoMapper.selectByCrossIdAndDirType(crossId, key);
            Integer isPedestrian = crossDirInfoPO.getIsPedestrian();
            dirListElement.setIsPersonCross(isPedestrian);
            // 构造内层List
            List<LaneInfoPO> value = entry.getValue();
            List<LaneListElement> laneListElementList = new ArrayList<>();
            for (LaneInfoPO laneInfoPO : value) {
                LaneListElement laneListElement = new LaneListElement();
                String laneId = laneInfoPO.getId();
                laneListElement.setId(laneId);
                laneListElement.setDirection(laneInfoPO.getTurn());
                laneListElement.setLaneType(laneInfoPO.getCategory());
                laneListElement.setName(laneInfoPO.getCode());
                // 根据laneId获取灯组代码
                String lightsCode = crossLaneLightsMapper.selectLightsCodeByLaneId(laneId);
                if (lightsCode != null) {
                    laneListElement.setLedNum(lightsCode);
                }
                laneListElementList.add(laneListElement);
            }
            dirListElement.setLaneList(laneListElementList);
            // 构造行人过街数据
            List<CrossLightsPO> crossLightsPOList = crossLightsMapper.selectByCrossIdAndDir(crossId, key);
            for (CrossLightsPO crossLightsPO : crossLightsPOList) {
                Integer type = crossLightsPO.getType();
                String lightsNo = crossLightsPO.getLightsNo();
                if (type == 20) { // 一次过街
                    dirListElement.setPersonCrossType("oneCross");
                    dirListElement.setOneCross(lightsNo);
                } else if (type == 22) { // 二次过街出口灯
                    dirListElement.setPersonCrossType("twiceCross");
                    dirListElement.setTwiceCrossOut(lightsNo);
                } else if (type == 21) { // 二次过街进口灯
                    dirListElement.setPersonCrossType("twiceCross");
                    dirListElement.setTwiceCrossIn(lightsNo);
                }
            }
            List<DirListElement> dirList = saveLaneInfoDTO.getDirList();
            dirList.add(dirListElement);
        }
    }

    private static String getId(String crossId, DirListElement dirListElement) {
        Integer dir = dirListElement.getDir();
        String id = crossId + "_" + dir + "_" + 1 + "_" + 0;// todo 暂无主辅路序号，赋值0
        return id;
    }

    private static CrossDirInfoPO getCrossDirInfoPO(String crossId, DirListElement dirListElement) {
        CrossDirInfoPO crossDirInfoPO = new CrossDirInfoPO();
        Integer dir = dirListElement.getDir();
        crossDirInfoPO.setId(crossId + "_" + dir + "_" + 1 + "_" + 0); // todo 暂无主辅路序号，赋值0
        crossDirInfoPO.setDirType(dir);
        crossDirInfoPO.setInOutType(1);
        crossDirInfoPO.setCrossId(crossId);
        crossDirInfoPO.setLength(0D); // todo 暂无长度，赋值0
        crossDirInfoPO.setIsPedestrian(dirListElement.getIsPersonCross());
        return crossDirInfoPO;
    }

    private static List<LaneInfoPO> getLaneInfoPOListForInsert(String crossId, DirListElement dirListElement,
                                                               Integer dir, List<LaneListElement> laneListFromClient) {
        List<LaneInfoPO> laneInfoPOListForInsert = new ArrayList<>();
        for (LaneListElement laneListElement : laneListFromClient) {
            LaneInfoPO laneInfoPO = new LaneInfoPO();
            // 序号
            String name = laneListElement.getName(); // 01
            String substring = name.substring(name.length() - 1);// 1
            String s = 1 + substring;
            int sort = Integer.parseInt(s);
            laneInfoPO.setId(crossId + "_" + dirListElement.getDir() + "_" + 2 + "_" + sort); // 信控车道类型都是进口车道
            laneInfoPO.setCode(name);
            laneInfoPO.setSort(sort);
            laneInfoPO.setDir(dir);
            laneInfoPO.setTurn(laneListElement.getDirection());
            laneInfoPO.setCategory(laneListElement.getLaneType());
            laneInfoPO.setCrossId(crossId);
            laneInfoPOListForInsert.add(laneInfoPO);
        }
        return laneInfoPOListForInsert;
    }

    private List<String> getLaneIds(List<LaneInfoPO> laneInfoPOList) {
        List<String> laneIds = new ArrayList<>();
        for (LaneInfoPO laneInfoPO : laneInfoPOList) {
            String id = laneInfoPO.getId();
            laneIds.add(id);
        }
        return laneIds;
    }

    private List<LaneListElement> buildLaneListFromDb(List<LaneInfoPO> laneInfoPOList) {
        List<LaneListElement> laneListElements = new ArrayList<>();
        for (LaneInfoPO laneInfoPO : laneInfoPOList) {
            LaneListElement laneListElement = new LaneListElement();
            laneListElement.setDirection(laneInfoPO.getTurn());
            laneListElement.setName(laneInfoPO.getCode());
            laneListElement.setLaneType(laneInfoPO.getCategory());
            laneListElements.add(laneListElement);
        }
        return laneListElements;
    }
}
